; this file contains code shared between `djula' and `cl-terrace' for interoperability
; or conveneince

(defpackage :djula-util
  (:use :cl :f-underscore)
  (:export ; utf-8 files
	   :slurp-utf-8-file

	   ; html/url encoding
	   :html-escape
	   :url-encode
	   :url-encode-path
	   :url-decode

	   ; filename / foldername
	   :filename
	   :foldername
	   
           ; boring files/directories
	   :*boring-path-regexps*
	   :boring-file-p
	   :non-boring-files
	   :boring-folder-p
	   :non-boring-folders

	   ; file function cache
	   :create-file-function-cache
	   :clear-file-function-write-date-cache
	   :sync-file-function-cache
	   :get-file-function))

(in-package :djula-util)

; configure these

(defparameter *boring-path-regexps* (list "^\\." "^\\_" "^#" "~$" ".fasl$"))

; utf-8 files

(defun slurp-utf-8-file (p)
  "return the contents of the file pointed to by the pathname designator `P' as a string assumes the file is encoded in :utf-8"
  (with-open-file (in p :element-type '(unsigned-byte 8)
		        :direction :input)
    (trivial-utf-8:read-utf-8-string in :stop-at-eof t)))

; url/html encoding

(defmacro .do-array ((var array &optional return) &body body)
  "like DOLIST but for arrays"
  (let ((n% (gensym "array-index"))
	(a%  (gensym "array")))
    `(let ((,a% ,array))
       (declare (dynamic-extent ,a%))
       (dotimes (,n% (length ,a%) ,return)
	 (let ((,var (aref ,a% ,n%)))
	   ,@body)))))

(defmacro .do-string ((var string &optional return) &body body)
  "like DOLIST but for strings"
  `(.do-array (,var (the string ,string) ,return)
     (declare (type character ,var))
     ,@body))

(defun html-escape (string)
  (with-output-to-string (out)
  "returns an html-escaped version of `STRING'"
  ;; this function was created by modifying HUNCHENTOOT:ESCAPE-FOR-HTML
  (.do-string (c string)
    (case c
      ((#\<) (write-string "&lt;" out))
      ((#\>) (write-string "&gt;" out))
      ((#\") (write-string "&quot;" out))
      ((#\') (write-string "&#039;" out))
      ((#\&) (write-string "&amp;" out))
      (otherwise (write-char c out))))))

(defun url-encode (string)
  "returns a url-encoded version of `STRING'. assumes UTF-8 so be careful if seen around HUNCHENTOOT:URL-DECODE, which probably assumes LATIN-1 by default. see HUCHENTOOT:*DEFAULT-EXTERNAL-FORAT*"
  ;; this function was created by modifying HUNCHENTOOT:URL-ENCODE
  (with-output-to-string (out)
    (.do-string (c string)
      (if (or (char<= #\0 c #\9)
	      (char<= #\a c #\z)
	      (char<= #\A c #\Z)

	      ;; note that there's no comma in there - because of cookies

	      ;;       I don't know why there's no comma because of cookies, it's
	      ;;       what hunchentoot did so I copied it -nick

	      (find c "$-_.!*'()" :test #'char=))
	  (write-char c out)
	  (.do-array (b (trivial-utf-8:string-to-utf-8-bytes (string c)))
	    (format out "%~2,'0x" b))))))

(defun url-encode-path (path)
  (with-output-to-string (out)
    (destructuring-bind (abs/rel . dpath) (pathname-directory path)

      ; maybe initial "/"
      (case abs/rel
	(:absolute (write-char #\/ out))
	(otherwise nil))

      ; the directory path
      (mapc (f_ (write-string (url-encode _) out)
		(write-char #\/ out))
	    dpath)

      ; the name
      (write-string (url-encode (pathname-name path)) out)

      ; maybe type
      (if (pathname-type path)
	  (format out ".~A" (pathname-type path))))))

(defun url-decode (string)
  "returns the a new url-decoded version of `STRING'. assumes UTF-8 so be careful if seen around HUNCHENTOOT:URL-ENCODE, which probably assumes LATIN-1 by default. see HUCHENTOOT:*DEFAULT-EXTERNAL-FORAT*"
  ;; this function was created by modifying HUNCHENTOOT:URL-DECODE
  (let ((vector (make-array (length string)
                            :element-type '(unsigned-byte 8)
                            :fill-pointer 0))
	percent-p
	buff)
    (declare (dynamic-extent vector))
    (dotimes (n (length string) (trivial-utf-8:utf-8-bytes-to-string (make-array (length vector)
										 :element-type '(unsigned-byte 8)
										 :initial-contents vector)))
      (let ((c (char string n)))
	(cond (buff          (vector-push (parse-integer string
							 :start (1- n)
							 :end (1+ n)
							 :radix 16)
					  vector)
			     (setq buff nil))
	      (percent-p     (setq buff t
				   percent-p nil))
	      ((char= c #\%) (setq percent-p t))
	      (t             (vector-push (char-code (case c
						       ((#\+)     #\Space)
						       (otherwise c)))
					  vector)))))))

; filename / foldernames

(defun filename (file)
  "
   (filename \"/foo/bar/baz.txt\")

   -> \"baz.txt\"
"
  (if (pathname-type file)
      (format nil "~A.~A" (pathname-name file) (pathname-type file))
      (pathname-name file)))

(defun foldername (folder)
  "
   (foldername \"/foo/bar/baz/\")

   -> \"baz\"
"
  (first (last (pathname-directory folder))))

; boring files / folders

(defun boring-file-p (path)
  "see *BORING-PATH-REGEXPTS*"
  (let ((% (filename path)))
    (some (f_ (cl-ppcre:scan _ %))
	  *boring-path-regexps*)))

(defun non-boring-files (folder)
  "returns a list of all the non-boring files in `FOLDER'"
  (remove-if 'boring-file-p
	     (remove-if 'cl-fad:directory-pathname-p
			(cl-fad:list-directory folder))))

(defun boring-folder-p (path)
  "see *BORING-PATH-REGEXPTS*"
  (let ((% (foldername path)))
    (some (f_ (cl-ppcre:scan _ %))
	  *boring-path-regexps*)))

(defun non-boring-folders (folder)
  "returns a list of all the non-boring folders in `FOLDER'"
  (remove-if 'boring-file-p
	     (remove-if-not 'cl-fad:directory-pathname-p
			    (cl-fad:list-directory folder))))


; file function cache

(defstruct file-function-cache
  (write-date-ht (make-hash-table :test 'equal))
  (function-ht (make-hash-table :test 'equal))
  fn-factory
  test)

(defmethod print-object ((ffc file-function-cache) s)
  (format s
	  "#<file-function-cache ~A ~A ~A>"
	  (hash-table-count (file-function-cache-write-date-ht ffc))
	  (hash-table-count (file-function-cache-function-ht ffc))
	  (file-function-cache-fn-factory ffc)))

(defun create-file-function-cache (fn-factory &key test)
  (make-file-function-cache :fn-factory fn-factory
			    :test test))

(defun clear-file-function-write-date-cache (cache)
  (clrhash (file-function-cache-write-date-ht cache))
  cache)

(defun .touch (path cache)
  (setf (gethash (namestring path) (file-function-cache-write-date-ht cache))
	(file-write-date path)))

(defun .out-of-date-p (path cache)
  (let ((known-write-date (gethash (namestring path) (file-function-cache-write-date-ht cache))))
    (or (not known-write-date)
	(> (file-write-date path) known-write-date))))

(defun .update-file-function (path cache)
  (let* ((string (namestring path))
	 (fn (funcall (file-function-cache-fn-factory cache)
		     string
		     (gethash string (file-function-cache-function-ht cache)))))

    ;; makes sure `FN-FACTORY' returned a function
    (if (not (functionp fn))
	(error "Cannot update the file-function-cache ~A for path ~A because ~A did not return a function"
	       (file-function-cache-fn-factory cache)
	       cache
	       path))

    ;; update `CACHE'
    (setf (gethash (namestring path) (file-function-cache-function-ht cache))
	  fn)
    (.touch path cache)
    fn))

(defun sync-file-function-cache (path cache)
  (when (.out-of-date-p path cache)
    (if (cl-fad:directory-pathname-p path)
	(mapc (f_ (sync-file-function-cache _ cache))
	      (nconc (non-boring-files path)
		     (non-boring-folders path)))
	(let ((test (file-function-cache-test cache)))
	  (if (or (not test)
		  (funcall test path))
	      (.update-file-function path cache))))))

(defun get-file-function (path cache &optional sync)
  (if sync
      (sync-file-function-cache path cache))
  (values (gethash (namestring path) (file-function-cache-function-ht cache))))